import pytest
from bloqade.ir.control import waveform, field, pulse, sequence
from bloqade.ir import analog_circuit
from bloqade.compiler.codegen.hardware.piecewise_linear import (
    PiecewiseLinear,
    GeneratePiecewiseLinearChannel,
)
from bloqade.compiler.codegen.hardware.piecewise_constant import (
    PiecewiseConstant,
    GeneratePiecewiseConstantChannel,
)
from bloqade.compiler.codegen.hardware.lattice_site_coefficients import (
    GenerateLatticeSiteCoefficients,
)
from bloqade import piecewise_linear, piecewise_constant, start
import numpy as np
from decimal import Decimal

from bloqade.ir.scalar import cast
from bloqade.submission.ir.parallel import ParallelDecoder, ClusterLocationInfo
from bloqade.ir.location import ListOfLocations, ParallelRegister
from bloqade.compiler.codegen.hardware.lattice import GenerateLattice
from bloqade.submission.capabilities import get_capabilities


@waveform.to_waveform(4.0)
def wf(t):
    return np.sin(t)


def test_error_pwl():
    with pytest.raises(TypeError):
        GeneratePiecewiseLinearChannel(
            sequence.rydberg, pulse.detuning, field.Uniform
        ).visit(wf)


def test_error_pwc():
    with pytest.raises(TypeError):
        GeneratePiecewiseConstantChannel(
            sequence.rydberg, pulse.detuning, field.Uniform
        ).visit(wf)


def test_waveform_append_pwl():
    visitor = GeneratePiecewiseLinearChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    wf = piecewise_linear(
        durations=[1, 2, 3],
        values=[0, 1, 0, 1],
    )

    expected = PiecewiseLinear(
        times=[Decimal("0"), Decimal("1"), Decimal("3"), Decimal("6")],
        values=[Decimal("0"), Decimal("1"), Decimal("0"), Decimal("1")],
    )
    assert visitor.visit(wf) == expected

    new_wf = wf.append(waveform.Poly([1, 3], 2)).append(waveform.Constant(7, 0.5))

    expected = PiecewiseLinear(
        times=[
            Decimal("0"),
            Decimal("1"),
            Decimal("3"),
            Decimal("6"),
            Decimal("8"),
            Decimal("8.5"),
        ],
        values=[
            Decimal("0"),
            Decimal("1"),
            Decimal("0"),
            Decimal("1"),
            Decimal("7"),
            Decimal("7"),
        ],
    )

    assert visitor.visit(new_wf) == expected


def test_waveform_append_pwc():
    visitor = GeneratePiecewiseConstantChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    wf1 = piecewise_constant(
        durations=[1, 2, 3],
        values=[1, 2, 3],
    )

    expected = PiecewiseConstant(
        [Decimal("0"), Decimal("1"), Decimal("3"), Decimal("6")],
        [Decimal("1"), Decimal("2"), Decimal("3"), Decimal("3")],
    )

    assert visitor.visit(wf1) == expected

    wf2 = wf1.append(waveform.Linear(5, 5, 1)).append(waveform.Poly([1], 4))

    expected = PiecewiseConstant(
        [
            Decimal("0"),
            Decimal("1"),
            Decimal("3"),
            Decimal("6"),
            Decimal("7"),
            Decimal("11"),
        ],
        [
            Decimal("1"),
            Decimal("2"),
            Decimal("3"),
            Decimal("5"),
            Decimal("1"),
            Decimal("1"),
        ],
    )

    assert visitor.visit(wf2) == expected


def test_waveform_sample_pwl():
    visitor = GeneratePiecewiseLinearChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    pwl_wf = wf.sample(0.1, waveform.Interpolation.Linear)

    times, values = pwl_wf.samples()

    expected = PiecewiseLinear(times=times, values=values)

    assert visitor.visit(pwl_wf) == expected


def test_waveform_sample_pwc():
    visitor = GeneratePiecewiseConstantChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    pwc_wf = wf.sample(0.1, waveform.Interpolation.Constant)

    times, values = pwc_wf.samples()

    values[-1] = values[-2]

    expected = PiecewiseConstant(times=times, values=values)

    assert visitor.visit(pwc_wf) == expected


def test_waveform_add_pwl():
    left = waveform.Linear(0, 1, 2)
    right = piecewise_linear([1, 1], [0, 1, 0])
    wf = left + right

    visitor = GeneratePiecewiseLinearChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseLinear(
        times=[Decimal("0"), Decimal("1"), Decimal("2")],
        values=[Decimal("0"), Decimal("1.5"), Decimal("1.0")],
    )

    assert visitor.visit(wf) == expected


def test_waveform_add_pwc():
    left = piecewise_constant([0.5, 1.5], [1, 0])
    right = piecewise_constant([1, 1], [0, 1])
    wf = left + right

    visitor = GeneratePiecewiseConstantChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseConstant(
        times=[Decimal("0.0"), Decimal("0.5"), Decimal("1.0"), Decimal("2")],
        values=[Decimal("1.0"), Decimal("0.0"), Decimal("1.0"), Decimal("1.0")],
    )

    assert visitor.visit(wf) == expected

    left = piecewise_constant([1, 1], [1, 0])
    right = piecewise_constant([1, 1], [0, 1])
    wf = left + right

    expected = PiecewiseConstant(
        times=[Decimal("0.0"), Decimal("1.0"), Decimal("2")],
        values=[Decimal("1.0"), Decimal("1.0"), Decimal("1.0")],
    )

    assert visitor.visit(wf) == expected


def test_waveform_slice_pwl():
    wf1 = waveform.Linear(0, 1, 2)[0.5:1.5]

    visitor = GeneratePiecewiseLinearChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseLinear(
        times=[Decimal("0"), Decimal("1")],
        values=[Decimal("0.25"), Decimal("0.75")],
    )

    assert visitor.visit(wf1) == expected

    wf2 = piecewise_linear([1, 2, 3], [0, 1, 0, 1])[0.5:4.5]

    expected = PiecewiseLinear(
        [Decimal("0.0"), Decimal("0.5"), Decimal("2.5"), Decimal("4")],
        [Decimal("0.5"), Decimal("1.0"), Decimal("0.0"), Decimal("0.5")],
    )

    assert visitor.visit(wf2) == expected

    wf3 = waveform.Linear(0, 1, 2)[0.5:0.5]

    expected = PiecewiseLinear(
        [Decimal("0.0"), Decimal("0.0")], [Decimal("0.0"), Decimal("0.0")]
    )

    assert visitor.visit(wf3) == expected

    wf4 = piecewise_linear([1, 2, 3], [0, 1, 0, 1])[1.0:4.5]

    expected = PiecewiseLinear(
        [Decimal("0.0"), Decimal("2.0"), Decimal("3.5")],
        [Decimal("1.0"), Decimal("0.0"), Decimal("0.5")],
    )

    assert visitor.visit(wf4) == expected

    wf5 = piecewise_linear([1, 2, 3], [0, 1, 0, 1])[:3.0]

    expected = PiecewiseLinear(
        [Decimal("0.0"), Decimal("1.0"), Decimal("3.0")],
        [Decimal("0.0"), Decimal("1.0"), Decimal("0.0")],
    )
    assert visitor.visit(wf5) == expected

    wf5 = piecewise_linear([1, 2, 3], [0, 1, 0, 1])[0.5:3.0]

    expected = PiecewiseLinear(
        [Decimal("0.0"), Decimal("0.5"), Decimal("2.5")],
        [Decimal("0.5"), Decimal("1.0"), Decimal("0.0")],
    )
    assert visitor.visit(wf5) == expected


def test_waveform_slice_pwc():
    visitor = GeneratePiecewiseConstantChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    wf1 = waveform.Constant(1, 2)[0.5:1.5]

    expected = PiecewiseConstant(
        times=[Decimal("0"), Decimal("1")],
        values=[Decimal("1"), Decimal("1")],
    )

    assert visitor.visit(wf1) == expected

    wf2 = piecewise_constant([1, 2, 3], [1, 2, 3])[0.5:4.5]

    expected = PiecewiseConstant(
        [Decimal("0.0"), Decimal("0.5"), Decimal("2.5"), Decimal("4")],
        [Decimal("1"), Decimal("2"), Decimal("3"), Decimal("3")],
    )

    assert visitor.visit(wf2) == expected

    wf3 = waveform.Constant(1, 2)[0.5:0.5]

    expected = PiecewiseConstant(
        [Decimal("0.0"), Decimal("0.0")], [Decimal("0"), Decimal("0")]
    )

    assert visitor.visit(wf3) == expected

    wf4 = piecewise_constant([1, 2, 3], [1, 2, 3])[1.0:4.5]

    expected = PiecewiseConstant(
        [Decimal("0.0"), Decimal("2.0"), Decimal("3.5")],
        [Decimal("2"), Decimal("3"), Decimal("3")],
    )

    assert visitor.visit(wf4) == expected

    wf5 = piecewise_constant([1, 2, 3], [1, 2, 3])[:3.0]

    expected = PiecewiseConstant(
        [Decimal("0.0"), Decimal("1.0"), Decimal("3.0")],
        [Decimal("1"), Decimal("2"), Decimal("2")],
    )

    assert visitor.visit(wf5) == expected

    wf5 = piecewise_constant([1, 2, 3], [1, 2, 3])[0.5:3.0]

    expected = PiecewiseConstant(
        [Decimal("0.0"), Decimal("0.5"), Decimal("2.5")],
        [Decimal("1"), Decimal("2"), Decimal("2")],
    )

    assert visitor.visit(wf5) == expected


def test_waveform_negative_pwl():
    wf = -waveform.Linear(0, 1, 2)

    visitor = GeneratePiecewiseLinearChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseLinear(
        times=[Decimal("0"), Decimal("2")],
        values=[Decimal("0"), Decimal("-1")],
    )

    assert visitor.visit(wf) == expected


def test_waveform_negative_pwc():
    wf = -piecewise_constant([1, 2, 3], [1, 2, 3])

    visitor = GeneratePiecewiseConstantChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseConstant(
        times=[Decimal("0"), Decimal("1"), Decimal("3"), Decimal("6")],
        values=[Decimal("-1"), Decimal("-2"), Decimal("-3"), Decimal("-3")],
    )

    assert visitor.visit(wf) == expected


def test_waveform_scale_pwl():
    wf = waveform.Linear(0, 1, 2) * 2

    visitor = GeneratePiecewiseLinearChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseLinear(
        times=[Decimal("0"), Decimal("2")],
        values=[Decimal("0"), Decimal("2")],
    )

    assert visitor.visit(wf) == expected


def test_waveform_scale_pwc():
    wf = piecewise_constant([1, 2, 3], [1, 2, 3]) * 2

    visitor = GeneratePiecewiseConstantChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseConstant(
        times=[Decimal("0"), Decimal("1"), Decimal("3"), Decimal("6")],
        values=[Decimal("2"), Decimal("4"), Decimal("6"), Decimal("6")],
    )

    assert visitor.visit(wf) == expected


def test_field_pwl():
    wf = piecewise_linear([1, 2, 3], [0, 1, 0, 1])
    f = field.Field({field.Uniform: wf})

    visitor = GeneratePiecewiseLinearChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseLinear(
        times=[Decimal("0"), Decimal("1"), Decimal("3"), Decimal("6")],
        values=[Decimal("0"), Decimal("1"), Decimal("0"), Decimal("1")],
    )

    assert visitor.visit(f) == expected


def test_field_pwc():
    wf = piecewise_constant([1, 2, 3], [1, 2, 3])
    f = field.Field({field.Uniform: wf})

    visitor = GeneratePiecewiseConstantChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseConstant(
        times=[Decimal("0"), Decimal("1"), Decimal("3"), Decimal("6")],
        values=[Decimal("1"), Decimal("2"), Decimal("3"), Decimal("3")],
    )

    assert visitor.visit(f) == expected


def test_pulse_leaf_pwl():
    wf = piecewise_linear([1, 2, 3], [0, 1, 0, 1])
    f = field.Field({field.Uniform: wf})
    p = pulse.Pulse({pulse.detuning: f})

    visitor = GeneratePiecewiseLinearChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseLinear(
        times=[Decimal("0"), Decimal("1"), Decimal("3"), Decimal("6")],
        values=[Decimal("0"), Decimal("1"), Decimal("0"), Decimal("1")],
    )

    assert visitor.visit(p) == expected


def test_pulse_leaf_pwc():
    wf = piecewise_constant([1, 2, 3], [1, 2, 3])
    f = field.Field({field.Uniform: wf})
    p = pulse.Pulse({pulse.detuning: f})

    visitor = GeneratePiecewiseConstantChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseConstant(
        times=[Decimal("0"), Decimal("1"), Decimal("3"), Decimal("6")],
        values=[Decimal("1"), Decimal("2"), Decimal("3"), Decimal("3")],
    )

    assert visitor.visit(p) == expected


def test_pulse_named_pwl():
    wf = piecewise_linear([1, 2, 3], [0, 1, 1, 0])
    f = field.Field({field.Uniform: wf})
    p = pulse.NamedPulse("test", pulse.Pulse({pulse.detuning: f}))

    visitor = GeneratePiecewiseLinearChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseLinear(
        times=[Decimal("0"), Decimal("1"), Decimal("3"), Decimal("6")],
        values=[Decimal("0"), Decimal("1"), Decimal("1"), Decimal("0")],
    )

    assert visitor.visit(p) == expected


def test_pulse_named_pwc():
    wf = piecewise_constant([1, 2, 3], [1, 2, 3])
    f = field.Field({field.Uniform: wf})
    p = pulse.NamedPulse("test", pulse.Pulse({pulse.detuning: f}))

    visitor = GeneratePiecewiseConstantChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseConstant(
        times=[Decimal("0"), Decimal("1"), Decimal("3"), Decimal("6")],
        values=[Decimal("1"), Decimal("2"), Decimal("3"), Decimal("3")],
    )

    assert visitor.visit(p) == expected


def test_pulse_slice_pwl():
    wf = piecewise_linear([1, 2, 3], [0, 1, 0, 1])
    f = field.Field({field.Uniform: wf})
    p = pulse.Pulse({pulse.detuning: f})[1.0:4.5]

    visitor = GeneratePiecewiseLinearChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseLinear(
        [Decimal("0.0"), Decimal("2.0"), Decimal("3.5")],
        [Decimal("1.0"), Decimal("0.0"), Decimal("0.5")],
    )

    assert visitor.visit(p) == expected


def test_pulse_slice_pwc():
    wf = piecewise_constant([1, 2, 3], [1, 2, 3])
    f = field.Field({field.Uniform: wf})
    p = pulse.Pulse({pulse.detuning: f})[1.0:4.5]

    visitor = GeneratePiecewiseConstantChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseConstant(
        [Decimal("0.0"), Decimal("2.0"), Decimal("3.5")],
        [Decimal("2"), Decimal("3"), Decimal("3")],
    )

    assert visitor.visit(p) == expected


def test_pulse_append_pwl():
    wf = piecewise_linear([1, 2, 3], [0, 1, 1, 0])
    f = field.Field({field.Uniform: wf})
    p = pulse.Pulse({pulse.detuning: f})
    p = p.append(p)

    visitor = GeneratePiecewiseLinearChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseLinear(
        [
            Decimal("0.0"),
            Decimal("1.0"),
            Decimal("3.0"),
            Decimal("6.0"),
            Decimal("7.0"),
            Decimal("9.0"),
            Decimal("12.0"),
        ],
        [
            Decimal("0.0"),
            Decimal("1.0"),
            Decimal("1.0"),
            Decimal("0.0"),
            Decimal("1.0"),
            Decimal("1.0"),
            Decimal("0.0"),
        ],
    )

    assert visitor.visit(p) == expected


def test_pulse_append_pwc():
    wf = piecewise_constant([1, 2, 3], [1, 2, 3])
    f = field.Field({field.Uniform: wf})
    p = pulse.Pulse({pulse.detuning: f})
    p = p.append(p)

    visitor = GeneratePiecewiseConstantChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseConstant(
        [
            Decimal("0.0"),
            Decimal("1.0"),
            Decimal("3.0"),
            Decimal("6.0"),
            Decimal("7.0"),
            Decimal("9.0"),
            Decimal("12.0"),
        ],
        [
            Decimal("1"),
            Decimal("2"),
            Decimal("3"),
            Decimal("1"),
            Decimal("2"),
            Decimal("3"),
            Decimal("3"),
        ],
    )

    assert visitor.visit(p) == expected


def test_sequence_leaf_pwl():
    wf = piecewise_linear([1, 2, 3], [0, 1, 0, 1])
    f = field.Field({field.Uniform: wf})
    p = pulse.Pulse({pulse.detuning: f})
    s = sequence.Sequence({sequence.rydberg: p})

    visitor = GeneratePiecewiseLinearChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseLinear(
        times=[Decimal("0"), Decimal("1"), Decimal("3"), Decimal("6")],
        values=[Decimal("0"), Decimal("1"), Decimal("0"), Decimal("1")],
    )

    assert visitor.visit(s) == expected


def test_sequence_leaf_pwc():
    wf = piecewise_constant([1, 2, 3], [1, 2, 3])
    f = field.Field({field.Uniform: wf})
    p = pulse.Pulse({pulse.detuning: f})
    s = sequence.Sequence({sequence.rydberg: p})

    visitor = GeneratePiecewiseConstantChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseConstant(
        times=[Decimal("0"), Decimal("1"), Decimal("3"), Decimal("6")],
        values=[Decimal("1"), Decimal("2"), Decimal("3"), Decimal("3")],
    )

    assert visitor.visit(s) == expected


def test_sequence_named_pwl():
    wf = piecewise_linear([1, 2, 3], [0, 1, 1, 0])
    f = field.Field({field.Uniform: wf})
    p = pulse.Pulse({pulse.detuning: f})
    s = sequence.NamedSequence("test", sequence.Sequence({sequence.rydberg: p}))

    visitor = GeneratePiecewiseLinearChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseLinear(
        times=[Decimal("0"), Decimal("1"), Decimal("3"), Decimal("6")],
        values=[Decimal("0"), Decimal("1"), Decimal("1"), Decimal("0")],
    )

    assert visitor.visit(s) == expected


def test_sequence_named_pwc():
    wf = piecewise_constant([1, 2, 3], [1, 2, 3])
    f = field.Field({field.Uniform: wf})
    p = pulse.Pulse({pulse.detuning: f})
    s = sequence.NamedSequence("test", sequence.Sequence({sequence.rydberg: p}))

    visitor = GeneratePiecewiseConstantChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseConstant(
        times=[Decimal("0"), Decimal("1"), Decimal("3"), Decimal("6")],
        values=[Decimal("1"), Decimal("2"), Decimal("3"), Decimal("3")],
    )

    assert visitor.visit(s) == expected


def test_sequence_slice_pwl():
    wf = piecewise_linear([1, 2, 3], [0, 1, 0, 1])
    f = field.Field({field.Uniform: wf})
    p = pulse.Pulse({pulse.detuning: f})
    s = sequence.Sequence({sequence.rydberg: p})[1.0:4.5]

    visitor = GeneratePiecewiseLinearChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseLinear(
        [Decimal("0.0"), Decimal("2.0"), Decimal("3.5")],
        [Decimal("1.0"), Decimal("0.0"), Decimal("0.5")],
    )

    assert visitor.visit(s) == expected


def test_sequence_slice_pwc():
    wf = piecewise_constant([1, 2, 3], [1, 2, 3])
    f = field.Field({field.Uniform: wf})
    p = pulse.Pulse({pulse.detuning: f})
    s = sequence.Sequence({sequence.rydberg: p})[1.0:4.5]

    visitor = GeneratePiecewiseConstantChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseConstant(
        [Decimal("0.0"), Decimal("2.0"), Decimal("3.5")],
        [Decimal("2"), Decimal("3"), Decimal("3")],
    )

    assert visitor.visit(s) == expected


def test_sequence_append_pwl():
    wf = piecewise_linear([1, 2, 3], [0, 1, 1, 0])
    f = field.Field({field.Uniform: wf})
    p = pulse.Pulse({pulse.detuning: f})
    s = sequence.Sequence({sequence.rydberg: p})
    s = s.append(s)

    visitor = GeneratePiecewiseLinearChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseLinear(
        [
            Decimal("0.0"),
            Decimal("1.0"),
            Decimal("3.0"),
            Decimal("6.0"),
            Decimal("7.0"),
            Decimal("9.0"),
            Decimal("12.0"),
        ],
        [
            Decimal("0.0"),
            Decimal("1.0"),
            Decimal("1.0"),
            Decimal("0.0"),
            Decimal("1.0"),
            Decimal("1.0"),
            Decimal("0.0"),
        ],
    )

    assert visitor.visit(s) == expected


def test_sequence_append_pwc():
    wf = piecewise_constant([1, 2, 3], [1, 2, 3])
    f = field.Field({field.Uniform: wf})
    p = pulse.Pulse({pulse.detuning: f})
    s = sequence.Sequence({sequence.rydberg: p})
    s = s.append(s)

    visitor = GeneratePiecewiseConstantChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseConstant(
        [
            Decimal("0.0"),
            Decimal("1.0"),
            Decimal("3.0"),
            Decimal("6.0"),
            Decimal("7.0"),
            Decimal("9.0"),
            Decimal("12.0"),
        ],
        [
            Decimal("1"),
            Decimal("2"),
            Decimal("3"),
            Decimal("1"),
            Decimal("2"),
            Decimal("3"),
            Decimal("3"),
        ],
    )

    assert visitor.visit(s) == expected


def test_analog_circuit_pwl():
    wf = piecewise_linear([1, 2, 3], [0, 1, 0, 1])
    f = field.Field({field.Uniform: wf})
    p = pulse.Pulse({pulse.detuning: f})
    s = sequence.Sequence({sequence.rydberg: p})
    c = analog_circuit.AnalogCircuit(start.add_position((0, 0)), s)

    visitor = GeneratePiecewiseLinearChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseLinear(
        times=[Decimal("0"), Decimal("1"), Decimal("3"), Decimal("6")],
        values=[Decimal("0"), Decimal("1"), Decimal("0"), Decimal("1")],
    )

    assert visitor.visit(c) == expected


def test_analog_circuit_pwc():
    wf = piecewise_constant([1, 2, 3], [1, 2, 3])
    f = field.Field({field.Uniform: wf})
    p = pulse.Pulse({pulse.detuning: f})
    s = sequence.Sequence({sequence.rydberg: p})
    c = analog_circuit.AnalogCircuit(start.add_position((0, 0)), s)

    visitor = GeneratePiecewiseConstantChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    )

    expected = PiecewiseConstant(
        times=[Decimal("0"), Decimal("1"), Decimal("3"), Decimal("6")],
        values=[Decimal("1"), Decimal("2"), Decimal("3"), Decimal("3")],
    )

    assert visitor.visit(c) == expected


def test_lattice_site_coefficients_codegen():
    wf = piecewise_linear([1, 2, 3], [0, 1, 0, 1])

    sm1 = field.AssignedRunTimeVector(
        "mask", [Decimal("0"), Decimal("1"), Decimal("2")]
    )
    sm2 = field.ScaledLocations.create({1: cast(1), 2: cast(2)})

    f1 = field.Field({sm1: wf})
    f2 = field.Field({sm2: wf})

    p1 = pulse.Pulse({pulse.detuning: f1})
    p2 = pulse.Pulse({pulse.detuning: f2})

    s1 = sequence.Sequence({sequence.rydberg: p1})
    s2 = sequence.Sequence({sequence.rydberg: p2})

    c1 = analog_circuit.AnalogCircuit(start.add_position([(0, 0), (1, 1), (2, 2)]), s1)
    c2 = analog_circuit.AnalogCircuit(start.add_position([(0, 0), (1, 1), (2, 2)]), s2)

    mapping = [
        ClusterLocationInfo(
            cluster_index=(0, 0), global_location_index=0, cluster_location_index=0
        ),
        ClusterLocationInfo(
            cluster_index=(0, 0), global_location_index=1, cluster_location_index=1
        ),
        ClusterLocationInfo(
            cluster_index=(0, 0), global_location_index=2, cluster_location_index=2
        ),
        ClusterLocationInfo(
            cluster_index=(0, 0), global_location_index=3, cluster_location_index=0
        ),
        ClusterLocationInfo(
            cluster_index=(0, 0), global_location_index=4, cluster_location_index=1
        ),
        ClusterLocationInfo(
            cluster_index=(0, 0), global_location_index=5, cluster_location_index=2
        ),
    ]

    parallel_decoder = ParallelDecoder(mapping)

    assert GenerateLatticeSiteCoefficients().emit(c1) == [
        Decimal("0"),
        Decimal("1"),
        Decimal("2"),
    ]
    assert GenerateLatticeSiteCoefficients().emit(c2) == [
        Decimal("0"),
        Decimal("1"),
        Decimal("2"),
    ]

    assert GenerateLatticeSiteCoefficients(parallel_decoder).emit(c1) == [
        Decimal("0"),
        Decimal("1"),
        Decimal("2"),
        Decimal("0"),
        Decimal("1"),
        Decimal("2"),
    ]


def test_lattice():
    lattice = ListOfLocations().add_position((0, 0)).add_position((0, 6), filling=False)

    capabilities = get_capabilities()

    capabilities.capabilities.lattice.area.height = Decimal("13.0e-6")
    capabilities.capabilities.lattice.area.width = Decimal("13.0e-6")
    capabilities.capabilities.lattice.geometry.number_sites_max = 4

    circuit = analog_circuit.AnalogCircuit(lattice, sequence.Sequence({}))

    ahs_lattice_data = GenerateLattice(capabilities).emit(circuit)

    assert ahs_lattice_data.sites == [
        (Decimal("0.0"), Decimal("0.0")),
        (Decimal("0.0"), Decimal("6.0")),
    ]
    assert ahs_lattice_data.filling == [1, 0]
    assert ahs_lattice_data.parallel_decoder is None

    parallel_lattice = ParallelRegister(lattice, cast(5))

    with pytest.raises(ValueError):
        ahs_lattice_data = GenerateLattice().emit(parallel_lattice)

    ahs_lattice_data = GenerateLattice(capabilities).emit(parallel_lattice)

    assert ahs_lattice_data.sites == [
        (Decimal("0.0"), Decimal("0.0")),
        (Decimal("0.0"), Decimal("6.0")),
        (Decimal("5.0"), Decimal("0.0")),
        (Decimal("5.0"), Decimal("6.0")),
    ]
    assert ahs_lattice_data.filling == [1, 0, 1, 0]
    assert ahs_lattice_data.parallel_decoder == ParallelDecoder(
        [
            ClusterLocationInfo(
                cluster_index=(0, 0), global_location_index=0, cluster_location_index=0
            ),
            ClusterLocationInfo(
                cluster_index=(0, 0), global_location_index=1, cluster_location_index=1
            ),
            ClusterLocationInfo(
                cluster_index=(1, 0), global_location_index=2, cluster_location_index=0
            ),
            ClusterLocationInfo(
                cluster_index=(1, 0), global_location_index=3, cluster_location_index=1
            ),
        ]
    )



from bloqade.builder.typing import ScalarType
from bloqade.builder.start import ProgramStart
from bloqade.ir.scalar import Scalar, Literal, cast
from bloqade.ir.tree_print import Printer

from pydantic.dataclasses import dataclass
from beartype.typing import List, Tuple, Generator, Union, Optional
from beartype import beartype
from enum import Enum
from numpy.typing import NDArray
from bloqade.submission.ir.capabilities import QuEraCapabilities
from bloqade.visualization import get_atom_arrangement_figure
from bloqade.visualization import display_ir

from beartype.vale import Is
from typing import Annotated
from plum import dispatch
import plotext as pltxt
import sys
import numpy as np


def check_position_array(array):
    return (
        array.ndim == 2
        and (
            np.issubdtype(array.dtype, np.floating)
            or np.issubdtype(array.dtype, np.integer)
        )
        and array.shape[1] == 2
    )


def check_bool_array(array):
    return array.ndim == 1 and np.issubdtype(array.dtype, np.bool_)


PositionArray = Annotated[np.ndarray, Is[check_position_array]]
BoolArray = Annotated[np.ndarray, Is[check_bool_array]]


class SiteFilling(int, Enum):
    filled = 1
    vacant = 0

    def print_node(self) -> str:
        return self.name

    def children(self) -> List:
        return []


@dataclass(frozen=True)
class LocationInfo:
    position: Tuple[Scalar, Scalar]
    filling: SiteFilling

    @beartype
    @staticmethod
    def create(position: Tuple[ScalarType, ScalarType], filled: bool):
        if filled:
            filling = SiteFilling.filled
        else:
            filling = SiteFilling.vacant

        position = tuple(cast(ele) for ele in position)

        return LocationInfo(position, filling)

    def print_node(self) -> str:
        return f"Location: {self.filling.name}"

    def children(self) -> List[Scalar]:
        return {"x": self.position[0], "y": self.position[1]}


@dataclass(init=False)
class AtomArrangement(ProgramStart):
    def __str__(self) -> str:
        def is_literal(x):
            return isinstance(x, Literal)

        plot_results = all(
            is_literal(info.position[0]) and is_literal(info.position[1])
            for info in self.enumerate()
        )

        if plot_results:
            xs = [float(info.position[0].value) for info in self.enumerate()]
            ys = [float(info.position[1].value) for info in self.enumerate()]
            filling = [bool(info.filling.value) for info in self.enumerate()]
            xs_filled = [x for x, filled in zip(xs, filling) if filled]
            ys_filled = [y for y, filled in zip(ys, filling) if filled]
            xs_vacant = [x for x, filled in zip(xs, filling) if not filled]
            ys_vacant = [y for y, filled in zip(ys, filling) if not filled]

            pltxt.clear_figure()
            pltxt.limit_size(False, False)
            pltxt.plot_size(80, 24)
            pltxt.canvas_color("default")
            pltxt.axes_color("default")
            pltxt.ticks_color("white")
            pltxt.title("Atom Positions")
            pltxt.xlabel("x (um)")
            pltxt.ylabel("y (um)")

            marker = "."
            if sys.stdout.encoding.lower().startswith("utf"):
                marker = "dot"

            pltxt.scatter(xs_filled, ys_filled, color=(100, 55, 255), marker=marker)
            if len(xs_vacant) > 0:
                pltxt.scatter(
                    xs_vacant, ys_vacant, color="white", label="vacant", marker=marker
                )

            return pltxt.build()
        else:
            ph = Printer()
            ph.print(self)
            return ph.get_value()

    def print_node(self) -> str:
        return "AtomArrangement"

    def children(self) -> List[LocationInfo]:
        return list(self.enumerate())

    def enumerate(self) -> Generator[LocationInfo, None, None]:
        """enumerate all locations in the register."""
        raise NotImplementedError

    def figure(self, fig_kwargs=None, **assignments):
        """obtain a figure object from the atom arrangement."""
        return get_atom_arrangement_figure(self, fig_kwargs, **assignments)

    def show(self, **assignments) -> None:
        display_ir(self, assignments)

    def rydberg_interaction(self, **assignments) -> NDArray:
        """calculate the Rydberg interaction matrix.

        Args:
            **assignments: the values to assign to the variables in the register.

        Returns:
            NDArray: the Rydberg interaction matrix in the lower triangular form.

        """

        from bloqade.constants import RB_C6

        # calculate the Interaction matrix
        V_ij = np.zeros((self.n_sites, self.n_sites))
        for i, site_i in enumerate(self.enumerate()):
            pos_i = np.array([float(ele(**assignments)) for ele in site_i.position])

            for j, site_j in enumerate(self.enumerate()):
                if j >= i:
                    break  # enforce lower triangular form

                pos_j = np.array([float(ele(**assignments)) for ele in site_j.position])
                r_ij = np.linalg.norm(pos_i - pos_j)

                V_ij[i, j] = RB_C6 / r_ij**6

        return V_ij

    @property
    def n_atoms(self) -> int:
        """number of atoms (filled sites) in the register."""
        raise NotImplementedError

    @property
    def n_sites(self) -> int:
        """number of sites in the register."""
        raise NotImplementedError

    @property
    def n_vacant(self) -> int:
        """number of vacant sites in the register."""
        raise NotImplementedError

    @property
    def n_dims(self) -> int:
        """number of dimensions in the register."""
        raise NotImplementedError

    @beartype
    def scale(self, scale: ScalarType):
        """
        Scale the geometry of your atoms.

        ### Usage Example:
        ```
        >>> reg = start.add_position([(0,0), (1,1)])
        # atom positions are now (0,0), (2,2)
        >>> new_reg = reg.scale(2)
        # you may also use scale on pre-defined geometries
        >>> from bloqade.atom_arrangement import Chain
        # atoms in the chain will now be 2 um apart versus
        # the default 1 um
        >>> Chain(11).scale(2)
        ```

        - Next possible steps are:
        - Continuing to build your geometry via:
            - `...add_position(positions).add_position(positions)`:
                to add more positions
            - `...add_position(positions).apply_defect_count(n_defects)`:
            to randomly drop out n_atoms
            - `...add_position(positions).apply_defect_density(defect_probability)`:
            to drop out atoms with a certain probability
            - `...add_position(positions).scale(scale)`: to scale the geometry
        - Targeting a level coupling once you're done with the atom geometry:
            - `...add_position(positions).rydberg`:
            to specify Rydberg coupling
            - `...add_position(positions).hyperfine`:
            to specify Hyperfine coupling
        - Visualizing your atom geometry:
            - `...add_position(positions).show()`:
            shows your geometry in your web browser

        """

        scale = cast(scale)
        location_list = []
        for location_info in self.enumerate():
            x, y = location_info.position
            new_position = (scale * x, scale * y)
            location_list.append(
                LocationInfo.create(new_position, bool(location_info.filling.value))
            )

        return ListOfLocations(location_list)

    @dispatch
    def _add_position(
        self, position: Tuple[ScalarType, ScalarType], filling: Optional[bool] = None
    ):
        if filling is None:
            filling = True

        location_list = list(self.enumerate())
        location_list.append(LocationInfo.create(position, filling))

        return ListOfLocations(location_list)

    @dispatch
    def _add_position(  # noqa: F811
        self,
        position: List[Tuple[ScalarType, ScalarType]],
        filling: Optional[List[bool]] = None,
    ):
        location_list = list(self.enumerate())

        assert (
            len(position) == len(filling) if filling else True
        ), "Length of positions and filling must be the same"

        if filling:
            for position, filling in zip(position, filling):
                location_list.append(LocationInfo.create(position, filling))

        else:
            for position in position:
                location_list.append(LocationInfo.create(position, True))

        return ListOfLocations(location_list)

    @dispatch
    def _add_position(  # noqa: F811
        self, position: PositionArray, filling: Optional[BoolArray] = None
    ):
        return self.add_position(
            list(map(tuple, position.tolist())),
            filling.tolist() if filling is not None else None,
        )

    def add_position(
        self,
        position: Union[
            PositionArray,
            List[Tuple[ScalarType, ScalarType]],
            Tuple[ScalarType, ScalarType],
        ],
        filling: Optional[Union[BoolArray, List[bool], bool]] = None,
    ) -> "ListOfLocations":
        """
        Add a position or multiple positions to a pre-existing geometry.

        `add_position` is capable of accepting:
        - A single tuple for one atom coordinate: `(1.0, 2.5)`
        - A list of tuples: `[(0.0, 1.0), (2.0,1.5), etc.]
        - A numpy array of shape (N, 2) where N is the number of atoms

        You may also intersperse variables anywhere a value may be present.

        You can also pass in an optional argument which determines the atom "filling"
        (whether or not at a specified coordinate an atom should be present).

        ### Usage Example:
        ```
        # single coordinate
        >>> reg = start.add_position((0,0))
        # you may chain add_position calls
        >>> reg_plus_two = reg.add_position([(2,2),(5.0, 2.1)])
        # you can add variables anywhere a value may be present
        >>> reg_with_var = reg_plus_two.add_position(("x", "y"))
        # and specify your atom fillings
        >>> reg_with_filling = reg_with_var.add_position([(3.1, 0.0), (4.1, 2.2)],
        [True, False])
        # alternatively you could use one boolean to specify
        # all coordinates should be empty/filled
        >>> reg_with_more_filling = reg_with_filling.add_positions([(3.1, 2.9),
        (5.2, 2.2)], False)
        ```

        - Next possible steps are:
        - Continuing to build your geometry via:
            - `...add_position(positions).add_position(positions)`:
                to add more positions
            - `...add_position(positions).apply_defect_count(n_defects)`:
            to randomly drop out n_atoms
            - `...add_position(positions).apply_defect_density(defect_probability)`:
            to drop out atoms with a certain probability
            - `...add_position(positions).scale(scale)`: to scale the geometry
        - Targeting a level coupling once you're done with the atom geometry:
            - `...add_position(positions).rydberg`: to specify Rydberg coupling
            - `...add_position(positions).hyperfine`: to specify Hyperfine coupling
        - Visualizing your atom geometry:
            - `...add_position(positions).show()`:
            shows your geometry in your web browser

        """
        return self._add_position(position, filling)

    @beartype
    def apply_defect_count(
        self, n_defects: int, rng: np.random.Generator = np.random.default_rng()
    ):
        """
        Drop `n_defects` atoms from the geometry randomly. Internally this occurs
        by setting certain sites to have a SiteFilling set to false indicating
        no atom is present at the coordinate.

        A default numpy-based Random Number Generator is used but you can
        explicitly override this by passing in your own.

        ### Usage Example:

        ```
        >>> from bloqade.atom_arrangement import Chain
        >>> import numpy as np
        # set a custom seed for a numpy-based RNG
        >>> custom_rng = np.random.default_rng(888)
        # randomly remove two atoms from the geometry
        >>> reg = Chain(11).apply_defect_count(2, custom_rng)
        # you may also chain apply_defect_count calls
        >>> reg.apply_defect_count(2, custom_rng)
        # you can also use apply_defect_count on custom geometries
        >>> from bloqade import start
        >>> start.add_position([(0,0), (1,1)]).apply_defect_count(1, custom_rng)
        ```

        - Next possible steps are:
        - Continuing to build your geometry via:
            - `...apply_defect_count(defect_counts).add_position(positions)`:
                to add more positions
            - `...apply_defect_count(defect_counts)
                .apply_defect_count(n_defects)`: to randomly drop out n_atoms
            - `...apply_defect_count(defect_counts)
                .apply_defect_density(defect_probability)`:
                to drop out atoms with a certain probability
            - `...apply_defect_count(defect_counts).scale(scale)`:
                to scale the geometry
        - Targeting a level coupling once you're done with the atom geometry:
            - `...apply_defect_count(defect_counts).rydberg`: to specify
                Rydberg coupling
            - `...apply_defect_count(defect_counts).hyperfine`:
                to specify Hyperfine coupling
        - Visualizing your atom geometry:
            - `...apply_defect_count(defect_counts).show()`:
                shows your geometry in your web browser
        """

        location_list = []
        for location_info in self.enumerate():
            location_list.append(location_info)

        filled_sites = []

        for index, location_info in enumerate(location_list):
            if location_info.filling is SiteFilling.filled:
                filled_sites.append(index)

        if n_defects >= len(filled_sites):
            raise ValueError(
                f"n_defects {n_defects} must be less than the number of filled sites "
                f"({len(filled_sites)})"
            )

        for _ in range(n_defects):
            index = rng.choice(filled_sites)
            location_list[index] = LocationInfo.create(
                location_list[index].position,
                (False if location_list[index].filling is SiteFilling.filled else True),
            )
            filled_sites.remove(index)

        return ListOfLocations(location_list)

    @beartype
    def apply_defect_density(
        self,
        defect_probability: float,
        rng: np.random.Generator = np.random.default_rng(),
    ):
        """
        Drop atoms randomly with `defect_probability` probability (range of 0 to 1).
        Internally this occurs by setting certain sites to have a SiteFilling
        set to false indicating no atom is present at the coordinate.

        A default numpy-based Random Number Generator is used but you can
        explicitly override this by passing in your own.

        ### Usage Example:

        ```
        >>> from bloqade.atom_arrangement import Chain
        >>> import numpy as np
        # set a custom seed for a numpy-based RNG
        >>> custom_rng = np.random.default_rng(888)
        # randomly remove two atoms from the geometry
        >>> reg = Chain(11).apply_defect_density(0.2, custom_rng)
        # you may also chain apply_defect_density calls
        >>> reg.apply_defect_count(0.1, custom_rng)
        # you can also use apply_defect_density on custom geometries
        >>> from bloqade import start
        >>> start.add_position([(0,0), (1,1)])
        .apply_defect_density(0.5, custom_rng)
        ```

        - Next possible steps are:
        - Continuing to build your geometry via:
            - `...apply_defect_count(defect_counts).add_position(positions)`:
            to add more positions
            - `...apply_defect_count(defect_counts).apply_defect_count(n_defects)`:
            to randomly drop out n_atoms
            - `...apply_defect_count(defect_counts)
            .apply_defect_density(defect_probability)`:
            to drop out atoms with a certain probability
            - `...apply_defect_count(defect_counts).scale(scale)`:
            to scale the geometry
        - Targeting a level coupling once you're done with the atom geometry:
            - `...apply_defect_count(defect_counts).rydberg`:
            to specify Rydberg coupling
            - `...apply_defect_count(defect_counts).hyperfine`:
            to specify Hyperfine coupling
        - Visualizing your atom geometry:
            - `...apply_defect_count(defect_counts).show()`:
            shows your geometry in your web browser
        """

        p = min(1, max(0, defect_probability))
        location_list = []

        for location_info in self.enumerate():
            if rng.random() < p:
                location_list.append(
                    LocationInfo.create(
                        location_info.position,
                        (
                            False
                            if location_info.filling is SiteFilling.filled
                            else True
                        ),
                    )
                )
            else:
                location_list.append(location_info)

        return ListOfLocations(location_list=location_list)

    def remove_vacant_sites(self):
        new_locations = []
        for location_info in self.enumerate():
            if location_info.filling is SiteFilling.filled:
                new_locations.append(location_info)

        return ListOfLocations(new_locations)


@dataclass
class ParallelRegister(ProgramStart):
    atom_arrangement: AtomArrangement
    cluster_spacing: Scalar

    @property
    def n_atoms(self):
        return self.atom_arrangement.n_atoms

    @property
    def n_sites(self):
        return self.atom_arrangement.n_sites

    @property
    def n_vacant(self):
        return self.atom_arrangement.n_vacant

    @property
    def n_dims(self):
        return self.atom_arrangement.n_dims

    def __str__(self):
        return "ParallelRegister:\n" + self.atom_arrangement.__str__()

    def _repr_pretty_(self, p, _):
        p.text(str(self))

    def _compile_to_list(
        self, __capabilities: Optional[QuEraCapabilities] = None, **assignments
    ):
        from bloqade.compiler.rewrite.common import AssignBloqadeIR
        from bloqade.compiler.codegen.hardware import GenerateLattice
        from bloqade.submission.capabilities import get_capabilities

        lattice_data = GenerateLattice(__capabilities or get_capabilities()).emit(
            AssignBloqadeIR(assignments).emit(self)
        )

        list_of_locations = ListOfLocations()
        for site, filling in zip(lattice_data.sites, lattice_data.filling):
            list_of_locations = list_of_locations.add_position(site, filling == 1)

        return list_of_locations

    def figure(
        self,
        fig_kwargs=None,
        capabilities: Optional[QuEraCapabilities] = None,
        **assignments,
    ):
        return self._compile_to_list(capabilities).figure(fig_kwargs)

    def show(self, **assignments) -> None:
        display_ir(self, assignments)


@dataclass(init=False)
class ParallelRegisterInfo:
    """ParallelRegisterInfo"""

    register_locations: List[List[Scalar]]
    register_filling: List[int]
    shift_vectors: List[List[Scalar]]

    def __init__(self, parallel_register: ParallelRegister):
        atom_arrangement = parallel_register.atom_arrangement
        cluster_spacing = parallel_register.cluster_spacing

        if atom_arrangement.n_atoms > 0:
            # calculate bounding box
            # of this register
            location_iter = atom_arrangement.enumerate()
            (x, y) = next(location_iter).position
            x_min = x
            x_max = x
            y_min = y
            y_max = y

            for location_info in location_iter:
                (x, y) = location_info.position
                x_min = x.min(x_min)
                x_max = x.max(x_max)
                y_min = y.min(y_min)
                y_max = y.max(y_max)

            shift_x = (x_max - x_min) + cluster_spacing
            shift_y = (y_max - y_min) + cluster_spacing

            register_locations = [
                list(location_info.position)
                for location_info in atom_arrangement.enumerate()
            ]
            register_filling = [
                location_info.filling.value
                for location_info in atom_arrangement.enumerate()
            ]
            shift_vectors = [[shift_x, cast(0)], [cast(0), shift_y]]
        else:
            raise ValueError("No locations to parallelize.")

        self.register_locations = register_locations
        self.register_filling = register_filling
        self.shift_vectors = shift_vectors


@dataclass(init=False)
class ListOfLocations(AtomArrangement):
    location_list: List[LocationInfo]

    @beartype
    def __init__(
        self,
        location_list: List[Union[LocationInfo, Tuple[ScalarType, ScalarType]]] = [],
    ):
        self.location_list = []
        for ele in location_list:
            if isinstance(ele, LocationInfo):
                self.location_list.append(ele)
            else:
                self.location_list.append(LocationInfo.create(ele, True))

        if self.location_list:
            self.__n_atoms = sum(
                1 for loc in self.location_list if loc.filling == SiteFilling.filled
            )
            self.__n_sites = len(self.location_list)
            self.__n_vacant = self.__n_sites - self.__n_atoms
            self.__n_dims = len(self.location_list[0].position)
        else:
            self.__n_sites = 0
            self.__n_atoms = 0
            self.__n_dims = None

        super().__init__()

    def _repr_pretty_(self, p, _):
        p.text(str(self))

    @property
    def n_atoms(self):
        return self.__n_atoms

    @property
    def n_sites(self):
        return self.__n_sites

    @property
    def n_vacant(self):
        return self.__n_vacant

    @property
    def n_dims(self):
        return self.__n_dims

    def enumerate(self):
        return iter(self.location_list)

    def __iter__(self):
        return iter(self.location_list)



import bloqade.ir.control.waveform as waveform
import bloqade.ir.control.field as field
import bloqade.ir.control.pulse as pulse
import bloqade.ir.control.sequence as sequence
from bloqade.ir.scalar import var

from bloqade.compiler.analysis.common.scan_channels import ScanChannels
from bloqade.compiler.rewrite.common.flatten import FlattenCircuit
from bloqade.compiler.rewrite.common.add_padding import AddPadding
from bloqade import tree_depth

tree_depth(30)


def test_flatten_pulses_simple():
    wf1 = waveform.Constant("v1", "t1")
    wf2 = waveform.Constant("v2", "t2")

    sm1 = field.UniformModulation()
    sm2 = field.RunTimeVector("mask1")

    f1 = field.Field({sm1: wf1})
    f2 = field.Field({sm2: wf2})

    p1 = pulse.Pulse({pulse.rabi.amplitude: f1})
    p2 = pulse.Pulse({pulse.rabi.phase: f2})

    t = var("t")
    p_test = p1.append(p2)[:t]
    interval = p_test.interval

    field_names = ScanChannels().scan(p_test)

    duration_1 = p1.duration
    duration_2 = p2.duration

    p_filled_expected = pulse.Slice(
        pulse.Append(
            [
                pulse.Pulse(
                    {
                        pulse.rabi.amplitude: f1,
                        pulse.rabi.phase: field.Field(
                            {sm2: waveform.Constant(0, duration_1)}
                        ),
                    }
                ),
                pulse.Pulse(
                    {
                        pulse.rabi.amplitude: field.Field(
                            {sm1: waveform.Constant(0, duration_2)}
                        ),
                        pulse.rabi.phase: f2,
                    }
                ),
            ]
        ),
        interval,
    )

    p_flattened_expected = pulse.Pulse(
        {
            pulse.rabi.amplitude: field.Field(
                {
                    sm1: wf1.append(waveform.Constant(0, duration_2))[:t],
                }
            ),
            pulse.rabi.phase: field.Field(
                {sm2: waveform.Constant(0, duration_1).append(wf2)[:t]}
            ),
        }
    )

    p_filled = AddPadding(field_names=field_names).visit(p_test)
    p_flattened = FlattenCircuit(field_names=field_names).visit(p_filled)

    assert p_filled_expected == p_filled
    assert p_flattened_expected == p_flattened


def test_flatten_pulses_complex():
    wf1 = waveform.Constant("v1", "t1")
    wf2 = waveform.Constant("v2", "t2")
    wf3 = waveform.Constant("v3", "t3")
    wf4 = waveform.Constant("v4", "t4")

    sm1 = field.UniformModulation()
    sm2 = field.RunTimeVector("mask1")
    sm3 = field.AssignedRunTimeVector("mask2", list(range(4)))
    sm4 = field.ScaledLocations.create(
        {field.Location(0): "m0", field.Location(2): "m2"}
    )

    f1 = field.Field({sm1: wf1})
    f2 = field.Field({sm2: wf2})
    f3 = field.Field({sm3: wf3, sm4: wf4})

    p1 = pulse.Pulse({pulse.rabi.amplitude: f1})
    p2 = pulse.Pulse({pulse.rabi.phase: f2, pulse.detuning: f3})

    t = var("t")
    p_test = p1.append(p2)[:t]
    interval = p_test.interval

    field_names = ScanChannels().scan(p_test)

    duration_1 = p1.duration
    duration_2 = p2.duration

    f2_padded = field.Field(
        {
            sm: wf.append(waveform.Constant(0, duration_2 - wf.duration))
            for sm, wf in f2.drives.items()
        }
    )
    f3_padded = field.Field(
        {
            sm: wf.append(waveform.Constant(0, duration_2 - wf.duration))
            for sm, wf in f3.drives.items()
        }
    )

    p_filled_expected = pulse.Slice(
        pulse.Append(
            [
                pulse.Pulse(
                    {
                        pulse.rabi.amplitude: f1,
                        pulse.rabi.phase: field.Field(
                            {
                                sm: waveform.Constant(0, duration_1)
                                for sm in f2.drives.keys()
                            }
                        ),
                        pulse.detuning: field.Field(
                            {
                                sm: waveform.Constant(0, duration_1)
                                for sm in f3.drives.keys()
                            }
                        ),
                    }
                ),
                pulse.Pulse(
                    {
                        pulse.rabi.amplitude: field.Field(
                            {sm1: waveform.Constant(0, duration_2)}
                        ),
                        pulse.rabi.phase: f2_padded,
                        pulse.detuning: f3_padded,
                    }
                ),
            ]
        ),
        interval,
    )

    p_flattened_expected = pulse.Pulse(
        {
            pulse.rabi.amplitude: field.Field(
                {sm1: wf1.append(waveform.Constant(0, duration_2))[:t]}
            ),
            pulse.rabi.phase: field.Field(
                {
                    sm: waveform.Constant(0, duration_1).append(wf)[:t]
                    for sm, wf in f2_padded.drives.items()
                }
            ),
            pulse.detuning: field.Field(
                {
                    sm: waveform.Constant(0, duration_1).append(wf)[:t]
                    for sm, wf in f3_padded.drives.items()
                }
            ),
        }
    )

    p_filled = AddPadding(field_names=field_names).visit(p_test)
    p_flattened = FlattenCircuit(field_names=field_names).visit(p_filled)

    assert p_filled_expected == p_filled
    assert p_flattened_expected == p_flattened


def test_flatten_sequence_simple():
    wf1 = waveform.Constant("v1", "t1")
    wf2 = waveform.Constant("v2", "t2")

    sm1 = field.UniformModulation()
    sm2 = field.RunTimeVector("mask1")

    f1 = field.Field({sm1: wf1})
    f2 = field.Field({sm2: wf2})

    p1 = pulse.Pulse({pulse.rabi.amplitude: f1})
    p2 = pulse.Pulse({pulse.rabi.phase: f2})

    s1 = sequence.Sequence({sequence.hyperfine: p1})
    s2 = sequence.Sequence({sequence.rydberg: p2})

    t = var("t")
    s_test = s1.append(s2)[:t]

    interval = s_test.interval

    p2_1 = pulse.Pulse(
        {pulse.rabi.phase: field.Field({sm2: waveform.Constant(0, p1.duration)})}
    )
    p1_2 = pulse.Pulse(
        {pulse.rabi.amplitude: field.Field({sm1: waveform.Constant(0, p2.duration)})}
    )

    s_filled_expected = sequence.Slice(
        interval=interval,
        sequence=sequence.Append(
            [
                sequence.Sequence({sequence.hyperfine: p1, sequence.rydberg: p2_1}),
                sequence.Sequence({sequence.hyperfine: p1_2, sequence.rydberg: p2}),
            ]
        ),
    )

    s_flattened_expected = sequence.Sequence(
        {
            sequence.hyperfine: pulse.Pulse(
                {
                    pulse.rabi.amplitude: field.Field(
                        {sm1: wf1.append(waveform.Constant(0, p2.duration))[:t]}
                    )
                },
            ),
            sequence.rydberg: pulse.Pulse(
                {
                    pulse.rabi.phase: field.Field(
                        {sm2: waveform.Constant(0, p1.duration).append(wf2)[:t]}
                    )
                }
            ),
        }
    )

    level_couplings = ScanChannels().scan(s_test)
    s_filled = AddPadding(level_couplings=level_couplings).visit(s_test)
    s_flattened = FlattenCircuit(level_couplings=level_couplings).visit(s_filled)
    print(s_filled_expected)
    print(s_filled)
    assert s_filled_expected == s_filled
    assert s_flattened_expected == s_flattened


def test_flatten_sequence_simple_2():
    wf1 = waveform.Constant("v1", "t1")
    wf2 = waveform.Constant("v2", "t2")

    sm1 = field.UniformModulation()
    sm2 = field.RunTimeVector("mask1")

    f1 = field.Field({sm1: wf1})
    f2 = field.Field({sm2: wf2})

    p1 = pulse.Pulse({pulse.rabi.amplitude: f1})
    p2 = pulse.Pulse({pulse.rabi.phase: f2})

    s_test = sequence.Sequence({sequence.hyperfine: p1, sequence.rydberg: p2})

    duration = s_test.duration

    p1_padding = pulse.Pulse(
        {
            pulse.rabi.amplitude: field.Field(
                {sm1: waveform.Constant(0, duration - p1.duration)}
            )
        }
    )
    p2_padding = pulse.Pulse(
        {
            pulse.rabi.phase: field.Field(
                {sm2: waveform.Constant(0, duration - p2.duration)}
            )
        }
    )

    s_filled_expected = sequence.Sequence(
        {
            sequence.hyperfine: p1.append(p1_padding),
            sequence.rydberg: p2.append(p2_padding),
        }
    )

    s_flattened_expected = sequence.Sequence(
        {
            sequence.hyperfine: pulse.Pulse(
                {
                    pulse.rabi.amplitude: field.Field(
                        {sm1: wf1.append(waveform.Constant(0, duration - p1.duration))}
                    )
                },
            ),
            sequence.rydberg: pulse.Pulse(
                {
                    pulse.rabi.phase: field.Field(
                        {sm2: wf2.append(waveform.Constant(0, duration - p2.duration))}
                    )
                }
            ),
        }
    )

    level_couplings = ScanChannels().scan(s_test)
    s_filled = AddPadding(level_couplings=level_couplings).visit(s_test)
    s_flattened = FlattenCircuit(level_couplings=level_couplings).visit(s_filled)

    assert s_filled_expected == s_filled
    assert s_flattened_expected == s_flattened
